# -*- coding: utf-8 -*-
## \namespace servidor
#  \brief Teacher's Bot

import logging
import logging.config
from threading import Lock
import base64
import os
import time
from hga.controlgrupos.common.exception import NoDefaultDevice

from hga.controlgrupos.common.sbotxmpp import SBotXMPP
from hga.controlgrupos.common.user_info import ping
from hga.controlgrupos.common.user_info import get_ip

from hga.controlgrupos.common.settings import LOGGING_CONFIG_FILE

from hga.controlgrupos.common.exception import NotImplementedException

from hga.controlgrupos.servidor.settingsservidor import RPC_ON_USER_CONNECT

from hga.controlgrupos.common.settings import DESKTOP_RUTA
from hga.controlgrupos.servidor.settingsservidor import RPC_MODULES

logging.config.fileConfig(LOGGING_CONFIG_FILE)
linfo = logging.info
lerror = logging.error
ldebug = logging.debug

## \class ServerProfesor(SBotXMPP)
#  \brief Teacher's Class
class ServerProfesor(SBotXMPP):
    ## \fn __init__(self, jid, muc_jid, nick, passwd="anonymous")
    #  \brief Inizialitation routines
    # \param jid The bot's JID.
    # \param muc_jid JID of the MUC  we are joining in.
    # \param nick Out bot's nick in the MUC.
    # \param passwd Password for the server. Not used.
    # \param parent Bot's parent object, normally the core.
    def __init__(self, jid, muc_jid, nick, passwd="anonymous", parent=None):
        self.RPC_MODULES = RPC_MODULES
        super(ServerProfesor, self).__init__(jid, muc_jid, nick, passwd, parent)
        # User list. JID as key, various as values :P
        self.users = {}
        # As events are async,. we don't want to mess with insertions an deletions
        # in users dict without protection :)
        self.users_lock = Lock()


    ## \fn  _load_event_handlers(self)
    #  \brief  Load event handlers 
    def _load_event_handlers(self):
        # As we are adding aditional event handlers, 
        # we need to let the parent add his own handlers
        super(ServerProfesor, self)._load_event_handlers()
        self.add_event_handler("groupchat_presence", self.handle_muc_presence)
        self.add_event_handler("groupchat_message", self.handle_message_recibed)

    ## \fn handle_muc_presence(self, entry)
    #  \brief Manages MUC presence events.
    # \param entry XMPP presence stanza
    def handle_muc_presence(self, entry):
        ldebug('Presence events\n%s\n', ('-'*80))
        user = entry.find('{%(ns)s}x/{%(ns)s}item' % 
            {'ns': 'http://jabber.org/protocol/muc#user'})
        user_muc_jid = entry.getFrom()
        nick = user_muc_jid.resource
        real_jid = user.get("jid")
        tipo = entry.getType()
        if real_jid == self.fulljid or real_jid is  None:
            # Our own presence or we have no permission to see the 
            # real JID of the user (shouldn't happen, this is bad).
            ldebug("Ignoring user: %s" % entry)
            return
        if tipo == "unavailable":
            # The user left the MUC.
            ldebug(u"Report to the core about user disappearing '%s'" % real_jid)
            self.core_handlers["DEL_USER_CB"](nick)
            return
        elif tipo == "available":
            # A user joins the MUC...
            if user.get("affiliation") != "member":
                # ... but we have to make him a member 
                # of the MUC before admitting him.
                self.muc.setAffiliation(room=self.muc_jid, 
                                        jid=real_jid, 
                                        affiliation="member")
                return

        # Finally, a full featured user :)
        linfo("User entered the MUC!: %s (%s)" % (nick, user_muc_jid))
        user_data = {'jid': real_jid, 'nick': nick, 'pending_rpc':{}}

        with self.users_lock:
            self.users[real_jid] = user_data
        # Run the default RPC calls on the user.
        self.process_default_rpc(real_jid)


    ## \fn process_default_rpc(self, jid)
    #  \brief Process default RPC on user connection
    #  \param jid User's JID
    def process_default_rpc(self, jid):
        linfo('Initialize RPC modules when user connect\n%s\n', ('-'*80))
        for rpc in RPC_ON_USER_CONNECT:
            id=self.rpc_call(jid, rpc, handler=self._default_rpc_handler)
            with self.users_lock:
                # We keep track of the pending RPC responses for the user.
                self.users[jid]["pending_rpc"][id] = rpc
        linfo('Sleekxmpp trace\n%s\n', ('-'*80))

    ## \fn  _default_rpc_handler(self, id, result)
    #  \brief  Handle RPC responses
    # \param id ID of the XMPP response stanza.
    # \param result Result of the RPC call.
    def _default_rpc_handler(self, id, result):
        linfo("Received RPC response\n%s\n" % ('-'*80))
        ldebug("RPC Response: id:'%s', result:'%s'" % (id, result))
        response_handled=False
        # List to know which users to delete once completed.
        # We need to track this because we can't modify the user list 
        # while iterating it.
        to_clean=[]
        with self.users_lock:
            for user in self.users:
                try:
                    call_name = self.users[user]["pending_rpc"].pop(id)
                    # Add the call name as key and result as value 
                    # to the user info dict.
                    self.users[user][call_name] = result
                    response_handled=True
                except KeyError, m:
                    ldebug("User '%s' not waiting for that response (%s): '%s'" % (user, id, result))
                if len(self.users[user]["pending_rpc"].keys()) == 0:
                    # No more pending RPCs for the user, report new user to Core...
                    self._report_new_user(self.users[user])
                    to_clean.append(user)

            for u in to_clean:
                # Delete all the users that got reported to the core.
                del self.users[u]

        if not response_handled:
            lerror("Unexpected RPC response (%s): '%s'" % (id, result))

    ## \fn _report_new_user(self, user)
    # \brief Reports the new user to the core.
    # \param user Diccionario con la información del usuario a ser añadido.
    def _report_new_user(self, user):
        del user["pending_rpc"]
        ldebug("New user ready: %s" % user)
        self.core_handlers["NEW_USER_CB"](user)

    ## \fn _ser_muc_config(self, **config)
    # \brief Stablishes the MUC config according to the named parameters.
    # Example: _set_muc_config(inviteonly=False, membersonly=True)
    # \param config Named parameters.
    def _set_muc_config(self, **config):
        linfo("Settings MUC config: '%s'" % config)
        # Dictionary comprehension to form the configuration update dictionary.
        room_config = dict( ("muc#roomconfig_%s" % c, config[c]) for c in config)
        config_form = self.muc.getRoomConfig(self.muc_jid)
        values = config_form.getValues()
        # Se actualiza con la nueva configuracion del diccionario.
        values.update(room_config)
        config_form.setValues(values)
        self.muc.setRoomConfig(self.muc_jid, config_form)

    # \fn abrir_muc(self)
    # \brief Configura el MUC permitiendo el acceso a todos los usuarios ("publico").
    def abrirMuc(self):
        self._set_muc_config(membersonly=False)

    # \fn cerrar_muc(self)
    # \brief Se configura el MUC como privado.
    def cerrarMuc(self):
        self._set_muc_config(membersonly=True)
    
    # \fn destruir_muc(self)
    # \brief #Destruye el muc cuando el profesor cierra el grupo
    def destruirMuc(self):
        self.muc.destroy(self.muc_jid)

    # \fn expulsar_usuario(self, nick)
    # \brief Se expulsa temporalmente a un usuario del MUC.
    # \param nick Nick del usuario a expulsar.
    def expulsarAlumno(self, nick):
        self.muc.setRole(self.muc_jid, nick=nick, role="none")

    # \fn banear_alumno(self, jid)
    # \brief Banea un alumno del MUC.
    # \param jid The JID of the user to kick.
    def bloquearAlumno(self, jid):
        # El usuario pasa a estar "outcast"
        self.muc.setAffiliation(self.muc_jid, jid=jid, affiliation="outcast")

    # \fn desbanear_alumno(self, jid)
    # \brief Desbanea un alumno del grupo. Debe realizars con el jid y no con el nick.
    # \param jid JID del alumno a readmitir.
    def desbloquearAlumno(self, jid):
        # El usuario pasa a ser "member"
        self.muc.setAffiliation(self.muc_jid, jid=jid, affiliation="member")

    ## \fn def handle_message_incomming(self, msg):
    # \brief Funcion para recibir mensajes
    # \param msg Mensaje que recibe el usuario
    def handle_message_recibed(self, msg):
        cuerpoMsg = msg['body']
        if cuerpoMsg.startswith('IMG'):
            listCommand = cuerpoMsg.split("|")
            ruta = os.path.join(DESKTOP_RUTA,listCommand[1].strip())
            if not os.path.exists(ruta):
                os.mkdir(ruta)
            imgagenC = os.path.join(DESKTOP_RUTA,listCommand[1].strip(),"captura.jpg")
            fh = open(imgagenC, "wb")
            fh.write(listCommand[2].decode('base64'))
            fh.close()
            self.core_handlers["ACTUALIZA_CAPTURA"]()

        # Funcionalidades módulo Control Desktop
        if cuerpoMsg.startswith('DESKTOP'):
            msg = cuerpoMsg.split("|")
            if msg[1] == "VNC":
                self.vnc_handlers["ACCESO_REMOTO"]()
            if msg[1] == "EMISION_ALUGRUPO_ON":
                self.vnc_handlers["EMI_ALUGRUPO_ON"]()


        if cuerpoMsg.startswith('HC'):
            msg = cuerpoMsg.split("|")
            self.ce_handlers["STATUS_SERVERHC"](msg[1])

        # Funcionalidad para peticion de ayuda por parte del alumno
        if cuerpoMsg.startswith('AYUDA'):
            listCommand = cuerpoMsg.split("|")
            nickAlumnoSolicitante = listCommand[1]
            self.core_handlers["OFRECER_AYUDA_CLIENTE"](nickAlumnoSolicitante)

        # Funcionalidad para realizar la captura de escritorio del alumno
        if cuerpoMsg.startswith('ENVIO_ESCRITORIO'):
            self.vnc_handlers["RECIBE_IMAGEN_ESCRITORIO"]()

    ## \fn def mensajeAlumno(self, jid_destino, jid_from, msg):
    # \brief Funcion que envia mensaje desde bot
    #  \param ipUser IP del usuario al que se le envia el mensaje
    #  \param jid_destino Es el jid del usuario al que se le envia el mensaje (cliente)
    #  \param jid_from Es el jid del que envia el mensaje (servidor)
    #  \param msg Es el mensaje que se le envia al usuario (cliente)
    def mensajeAlumno(self, jid_destino, jid_from, msg):
        try:
            get_ip() # Si no hay conexión NoDefaultDevice
            self.sendMessage(mto=jid_destino, mbody=msg, mtype='groupchat', mfrom=jid_from)
        except KeyError, e:
            lerror("ServerProfesor -> mensajeAlumno: %s" % e)
        except NoDefaultDevice, e:
            lerror("ServerProfesor -> mensajeAlumno: %s" % e)
            
    # \fn set_core_handlers(self, handlers)
    # \brief Set the handlers for the different events the bot can receive.
    # \param handlers Dictionary in the form EVENT:handler
    def set_core_handlers(self, handlers):
        self.core_handlers = handlers

    def vnc_handlers(self, handlers):
        self.vnc_handlers = handlers

    def set_ce_handlers(self, handlers):
        self.ce_handlers = handlers

if __name__=="__main__":
    host = "monkey.cga.andared.cec.junta-andalucia.es"
    jid = "@%s" % (host)
    password = "profesor"
    muc_jid = "salita@conference.%s" % (host)
    nick = "Un bot de profesor cualquiera"
    bot = ServerProfesor(jid, muc_jid, nick, passwd=password)
    bot.connect()
    bot.process(threaded=False)
